package cc.fascinated.bat.service;

import cc.fascinated.bat.BatApplication;
import cc.fascinated.bat.common.DateUtils;
import cc.fascinated.bat.common.WebRequest;
import cc.fascinated.bat.common.beatsaber.leaderboard.impl.ScoreSaberLeaderboard;
import cc.fascinated.bat.event.EventListener;
import cc.fascinated.bat.exception.BadRequestException;
import cc.fascinated.bat.exception.ResourceNotFoundException;
import cc.fascinated.bat.features.scoresaber.profile.user.ScoreSaberProfile;
import cc.fascinated.bat.model.token.beatsaber.scoresaber.*;
import com.google.gson.JsonObject;
import lombok.*;
import lombok.extern.log4j.Log4j2;
import net.jodah.expiringmap.ExpiringMap;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.DependsOn;
import org.springframework.stereotype.Service;
import org.springframework.web.socket.CloseStatus;
import org.springframework.web.socket.TextMessage;
import org.springframework.web.socket.WebSocketSession;
import org.springframework.web.socket.client.standard.StandardWebSocketClient;
import org.springframework.web.socket.handler.TextWebSocketHandler;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeUnit;
import java.util.function.Consumer;

@Service
@Log4j2(topic = "ScoreSaber Service")
@DependsOn("discordService")
public class ScoreSaberService extends TextWebSocketHandler {
    private static final String SCORESABER_API = "https://scoresaber.com/api/";
    private static final String GET_PLAYER_ENDPOINT = SCORESABER_API + "player/%s/full";
    private static final String GET_PLAYER_SCORES_ENDPOINT = SCORESABER_API + "player/%s/scores?limit=100&sort=%s&page=%s&withMetadata=true";

    /**
     * The cached accounts.
     */
    private final Map<String, ScoreSaberAccountToken> cachedAccounts = ExpiringMap.builder()
            .expiration(5, TimeUnit.MINUTES)
            .build();

    /**
     * The cached score pages.
     */
    private final Map<String, CachedPage> cachedScorePages = ExpiringMap.builder()
            .expiration(30, TimeUnit.MINUTES)
            .build();

    @Autowired
    public ScoreSaberService() {
        connectWebSocket();
    }

    /**
     * Checks if the account is cached.
     *
     * @param id The id of the account.
     * @return If the account is cached.
     */
    public boolean isCached(String id) {
        return cachedAccounts.containsKey(id);
    }

    /**
     * Gets the account from the ScoreSaber API.
     *
     * @param id The id of the account.
     * @return The account.
     * @throws ResourceNotFoundException                      If the account is not found.
     * @throws cc.fascinated.bat.exception.RateLimitException If the ScoreSaber rate limit is reached.
     */
    public ScoreSaberAccountToken getAccount(String id) {
        if (cachedAccounts.containsKey(id)) {
            return cachedAccounts.get(id);
        }

        ScoreSaberAccountToken account = WebRequest.getAsEntity(String.format(GET_PLAYER_ENDPOINT, id), ScoreSaberAccountToken.class);
        if (account == null) { // Check if the account doesn't exist.
            log.info("Account with id '{}' not found.", id);
            throw new ResourceNotFoundException("Account with id '%s' not found.".formatted(id));
        }
        if (account.isBanned()) {
            throw new BadRequestException("Account with id '%s' is banned.".formatted(id));
        }
        if (account.isInactive()) {
            throw new BadRequestException("Account with id '%s' is inactive.".formatted(id));
        }
        cachedAccounts.put(id, account);
        return account;
    }

    /**
     * Gets the scores for the account.
     *
     * @param profile The profile.
     * @param page    The page to get the scores from.
     * @return The scores.
     */
    public CachedPage getPageScores(ScoreSaberProfile profile, int page) {
        String key = profile.getAccountId() + "." + page;
        // Check if the page is cached, but don't cache the first page as it's the most likely to change.
        if (cachedScorePages.containsKey(key) && page != 1) {
            CachedPage cachedPage = cachedScorePages.get(key);
            cachedPage.setCached(true);
            return cachedPage;
        }

        log.info("Fetching scores for account '{}' from page {}.", profile.getAccountId(), page);
        ScoreSaberScoresPageToken pageToken = WebRequest.getAsEntity(String.format(GET_PLAYER_SCORES_ENDPOINT, profile.getAccountId(), "recent", page), ScoreSaberScoresPageToken.class);
        if (pageToken == null) { // Check if the page doesn't exist.
            return null;
        }
        // Sort the scores by newest time set.
        pageToken.setPlayerScores(Arrays.stream(pageToken.getPlayerScores())
                .sorted((a, b) -> DateUtils.getDateFromString(b.getScore().getTimeSet()).compareTo(DateUtils.getDateFromString(a.getScore().getTimeSet())))
                .toArray(ScoreSaberPlayerScoreToken[]::new));
        CachedPage cachedPage = new CachedPage(pageToken, false);
        cachedScorePages.put(key, cachedPage);
        return cachedPage;
    }

    /**
     * Gets the scores for the account.
     *
     * @param profile The profile.
     * @return The scores.
     */
    public List<CachedPage> getScores(ScoreSaberProfile profile, Consumer<CurrentPageCallback> currentPageCallback) {
        List<CachedPage> scores = new ArrayList<>(List.of(getPageScores(profile, 1)));
        ScoreSaberPageMetadataToken metadata = scores.get(0).getPage().getMetadata();
        int totalPages = (int) Math.ceil((double) metadata.getTotal() / metadata.getItemsPerPage());
        log.info("Fetching {} pages of scores for account '{}'.", totalPages, profile.getAccountId());
        for (int i = 2; i <= totalPages; i++) {
            scores.add(getPageScores(profile, i));
            currentPageCallback.accept(new CurrentPageCallback(i, totalPages));
        }
        return scores;
    }

    /**
     * Gets the raw pp per global pp for the given account.
     *
     * @param account The account.
     * @return The raw pp per global pp.
     */
    public RawPerGlobal getRawPerGlobal(ScoreSaberProfile account, Consumer<CurrentPageCallback> currentPageCallback) {
        List<CachedPage> scores = getScores(account, currentPageCallback);
        List<ScoreSaberScoreToken> playerScores = new ArrayList<>();
        for (CachedPage score : scores) {
            for (ScoreSaberPlayerScoreToken playerScore : score.getPage().getPlayerScores()) {
                playerScores.add(playerScore.getScore());
            }
        }
        return new RawPerGlobal(
                ScoreSaberLeaderboard.INSTANCE.getRawPerGlobalPP(playerScores, 1),
                (int) scores.stream().filter(CachedPage::isCached).count(),
                scores.size()
        );
    }

    /**
     * Connects to the ScoreSaber WebSocket.
     */
    @SneakyThrows
    private void connectWebSocket() {
        log.info("Connecting to the ScoreSaber WSS.");
        new StandardWebSocketClient().execute(this, "wss://scoresaber.com/ws").get();
    }

    @Override
    public void afterConnectionEstablished(@NonNull WebSocketSession session) {
        log.info("Connected to the ScoreSaber WSS.");
    }

    @Override
    public void afterConnectionClosed(@NonNull WebSocketSession session, @NonNull CloseStatus status) {
        log.info("Disconnected from the ScoreSaber WSS.");
        connectWebSocket(); // Reconnect to the WebSocket.
    }

    @Override
    @SneakyThrows
    protected void handleTextMessage(@NonNull WebSocketSession session, @NonNull TextMessage message) {
        // Ignore the connection message.
        if (message.getPayload().equals("Connected to the ScoreSaber WSS")) {
            return;
        }

        try {
            JsonObject json = BatApplication.GSON.fromJson(message.getPayload(), JsonObject.class);
            String command = json.get("commandName").getAsString();
            JsonObject data = json.get("commandData").getAsJsonObject();

            if (command.equals("score")) {
                ScoreSaberPlayerScoreToken score = BatApplication.GSON.fromJson(data, ScoreSaberPlayerScoreToken.class);
                ScoreSaberScoreToken.LeaderboardPlayerInfo playerInfo = score.getScore().getLeaderboardPlayerInfo();

                for (EventListener listener : EventService.LISTENERS) {
                    listener.onScoresaberScoreReceived(score, score.getLeaderboard(), playerInfo);
                }
            }
        } catch (Exception ex) {
            log.error("An error occurred while handling the message.", ex);
        }
    }

    @AllArgsConstructor
    @Getter
    public static class CurrentPageCallback {
        private int currentPage;
        private int totalPages;
    }

    @AllArgsConstructor
    @Getter
    @Setter
    public static class CachedPage {
        /**
         * The page of scores.
         */
        private ScoreSaberScoresPageToken page;

        /**
         * Whether the page is cached.
         */
        private boolean cached;
    }

    @AllArgsConstructor
    @Getter
    public static class RawPerGlobal {
        /**
         * The raw pp per global pp.
         */
        private double rawPerGlobal;

        /**
         * The amount of pages that were cached.
         */
        private int cachedPages;

        /**
         * The total amount of pages.
         */
        private int totalPages;
    }
}
